#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

static char const prog[]="preconf_subst";

typedef int quoter(char const*, char const**, char* const*);

static quoter quoter_sh;
static quoter quoter_c;
static quoter quoter_sed;

typedef struct {
  char const* name;
  quoter* quoter;
} quote_type;

static quote_type const types[]={
  { "sh",   quoter_sh   },
  { "c",    quoter_c    },
  { "sed",  quoter_sed  }
};

int main(int argc, char** argv) {
  size_t i;
  char const* q;

  if (argc<3) {
    fputs(prog, stderr);
    fputs(usage, stderr);
    fputs(prog, stderr);
    fputs(" quote_type value", stderr);
    return 111;
  }

  for (i=0;; ++i) {
    if (i==sizeof types/sizeof types[0]) {
      fputs(prog, stderr);
      fputs(fatal, stderr);
      fputs("unrecognized quote type: ", stderr);
      fputs(argv[1], stderr);
      fputs("\n", stderr);
      return 111;
    }
    if (strcmp(argv[1], types[i].name)==0) break;
  }

  if (types[i].quoter(argv[2], &q, argv+3)!=0) {
    fputs(prog, stderr);
    fputs(fatal, stderr);
    fputs(q, stderr);
    fputs("\n", stderr);
    return 111;
  }
  fputs(q, stdout);
  fputs("\n", stdout);

  return 0;
}

static char const err_overflow[]="quoted length overflows size_t";

static int quoter_sh(char const* s, char const** out, char* const* args) {
  char* qs;
  size_t len, i;

  (void)args;

  for (len=2, i=0; s[i]!='\0'; ++i) {
    switch (s[i]) {
      case '\\': case '\'':
        len+=4;
        if (len<4) goto overflow;
        break;
      default:
        if (++len==0) goto overflow;
        break;
    }
  }
  if (++len==0) goto overflow;

  qs=malloc(len);
  if (qs==null) {
    *out=err_alloc;
    return errno;
  }

  qs[0]='\'';
  for (len=1, i=0; s[i]!='\0'; ++i) {
    switch (s[i]) {
      case '\\': case '\'':
        qs[len++]='\'';
        qs[len++]='\\';
        qs[len++]=s[i];
        qs[len++]='\'';
        break;
      default:
        qs[len++]=s[i];
        break;
    }
  }
  qs[len++]='\'';
  qs[len]='\0';

  *out=qs;
  return 0;
  overflow:
  *out=err_overflow;
  return -1;
}

static int quoter_c(char const* s, char const** out, char* const* args) {
  char* qs;
  size_t len, i;

  (void)args;

  i=strlen(s);
  len=4*i;
  if (len/4!=i) goto overflow;
  len+=3;
  if (len<3) goto overflow;

  qs=malloc(len);
  if (qs==null) {
    *out=err_alloc;
    return errno;
  }

  qs[0]='"';
  for (len=1, i=0; s[i]!='\0'; ++i) {
    qs[len++]='\\';
    qs[len++]='0'+((s[i]>>6)&07);
    qs[len++]='0'+((s[i]>>3)&07);
    qs[len++]='0'+(s[i]&07);
  }
  qs[len++]='"';
  qs[len]='\0';

  *out=qs;
  return 0;
  overflow:
  *out=err_overflow;
  return -1;
}

static int quoter_sed(char const* s, char const** out, char* const* args) {
  char const q=(args[0]==null? '/': args[0][0]);
  char* qs;
  size_t len, i;

  for (len=0, i=0; s[i]!='\0'; ++i) {
    if (++len==0) goto overflow;
    if (s[i]=='\\' || s[i]=='&' || s[i]==q)
      if (++len==0) goto overflow;
  }
  if (++len==0) goto overflow;

  qs=malloc(len);
  if (qs==null) {
    *out=err_alloc;
    return errno;
  }

  for (len=0, i=0; s[i]!='\0'; ++i) {
    if (s[i]=='\\' || s[i]=='&' || s[i]==q)
      qs[len++]='\\';
    qs[len++]=s[i];
  }
  qs[len]='\0';

  *out=qs;
  return 0;
  overflow:
  *out=err_overflow;
  return -1;
}
